#include <pthread.h>
#include <unistd.h>
#include <signal.h>
#include <poll.h>
#include <errno.h>
#include <stdio.h>
#include <sys/socket.h>
#include <linux/rtnetlink.h>
#include <netinet/in.h>
#include <netinet/icmp6.h>
#include <net/if.h>
#include <sys/eventfd.h>
#include "klink.h"
#include "manager.h"
#include "missing.h"

static int efd=-1;
static int kfd=-1;
static char *dev=NULL;
static pthread_t wh;
static unsigned char bfr[65536+8];

static void parse(struct nd_opt_hdr *opt,int len)
{
	int i;
	int optlen;
	int valid;
	struct nd_opt_rdnss *rdnss;
	struct in6_addr *a;
	NSLIST l;

	for (;len>=sizeof(struct nd_opt_hdr);len-=opt->nd_opt_len<<3,
	     opt=(struct nd_opt_hdr *)((uint8_t *)opt+(opt->nd_opt_len<<3)))
	{
		if (!opt->nd_opt_len||len<(opt->nd_opt_len<<3))return;

		if(opt->nd_opt_type!=ND_OPT_RDNSS)continue;
		if(opt->nd_opt_len<3||!(opt->nd_opt_len&1))continue;

		rdnss=(struct nd_opt_rdnss *)opt;
		optlen=opt->nd_opt_len;
		a=(struct in6_addr *)(rdnss+1);
		valid=ntohl(rdnss->nd_opt_rdnss_lifetime);

		for(l.total=0;optlen>=2&&l.total<MAX_NS;a++,optlen-=2)
		{
			for(i=0;i<l.total;i++)
			    if(!memcmp(&l.addr[i],a,sizeof(struct in6_addr)))
				break;
			if(i<l.total)continue;

			l.addr[l.total]=*a;
			l.valid[l.total++]=valid;
		}

		mgr_cmd(CMD_NSLIST,&l,sizeof(NSLIST));
	}
}

static void *worker(void *data)
{
	struct pollfd p[2];
	uint64_t dummy;
	sigset_t s;
	int size;
	int len;
	struct nduseroptmsg *msg;

	sigfillset(&s);
	pthread_sigmask(SIG_BLOCK,&s,NULL);

	p[0].fd=efd;
	p[0].events=POLLIN;
	p[1].fd=kfd;
	p[1].events=POLLIN;

	while(1)
	{
		if(poll(p,2,-1)<=0)continue;

		if(p[0].revents&POLLIN)
		{
			if(read(efd,&dummy,sizeof(dummy))!=sizeof(dummy))
				perror("read eventfd");
			break;
		}

		if(p[1].revents&POLLIN)
		{
			size=NLMSG_SPACE(65536-sizeof(struct icmp6_hdr));
			memset(bfr,0,size);

			if((len=recv(kfd,bfr,size,0))==-1)
			{
				if(errno==EAGAIN)continue;
				perror("recv AF_NETLINK");
				break;
			}

			if(len<NLMSG_SPACE(sizeof(struct nduseroptmsg)))
				continue;

			msg=(struct nduseroptmsg *)
				NLMSG_DATA((struct nlmsghdr *)bfr);

			if(msg->nduseropt_family!=AF_INET6||
				msg->nduseropt_icmp_type!=ND_ROUTER_ADVERT||
				msg->nduseropt_icmp_code)continue;

			if(msg->nduseropt_ifindex!=if_nametoindex(dev))
				continue;

			parse((struct nd_opt_hdr *)(msg+1),
				msg->nduseropt_opts_len);
		}
	}

	pthread_exit(NULL);
	return NULL;
}

void kl_exit(void)
{
	if(efd!=-1)close(efd);
	if(kfd!=-1)close(kfd);
	if(dev)free(dev);
	efd=-1;
	kfd=-1;
	dev=NULL;
}

int kl_init(char *ifc)
{
	struct sockaddr_nl a;

	if(dev)free(dev);
	dev=strdup(ifc);
	if(!dev)
	{
		perror("strdup");
		kl_exit();
		return -1;
	}

	if(efd==-1)
	{
		efd=eventfd(0,EFD_CLOEXEC);
		if(efd==-1)
		{
			perror("eventfd");
			kl_exit();
			return -1;
		}
	}

	if(kfd==-1)
	{
		kfd=socket(AF_NETLINK,SOCK_DGRAM|SOCK_CLOEXEC|SOCK_NONBLOCK,
			NETLINK_ROUTE);
		if(kfd==-1)
		{
			perror("socket AF_NETLINK NETLINK_ROUTE");
			kl_exit();
			return -1;
		}

		memset(&a,0,sizeof(a));
		a.nl_family=AF_NETLINK;
		a.nl_groups=1<<(RTNLGRP_ND_USEROPT-1);

		if(bind(kfd,(struct sockaddr *)&a,sizeof(struct sockaddr_nl)))
		{
			perror("bind AF_NETLINK");
			kl_exit();
			return -1;
		}
	}

	return 0;
}

int kl_start(void)
{
	if(pthread_create(&wh,NULL,worker,NULL))
	{
		perror("pthread_create");
		return -1;
	}

	return 0;
}

void kl_stop(void)
{
	uint64_t dummy=1;

	write(efd,&dummy,sizeof(dummy));
	pthread_join(wh,NULL);
}
