#include <unistd.h>
#include <signal.h>
#include <stdio.h>
#include <netinet/in.h>
#include "requester.h"
#include "klink.h"
#include "wireless.h"
#include "manager.h"
#include "output.h"

#define exit(a) do { fflush(stdout); fflush(stderr); _exit(a); } while(0)

static void dummy(int unused)
{
}

static int infi(int mode,char *dev,int burst,CFG *cfg,char *script,char *cfn)
{
	int err=-1;

	if(!mode)
	{
		err=0;
		goto out;
	}

	if(mgr_init(burst,cfg))
	{
		fprintf(stderr,"mgr_init failure\n");
		goto out;
	}

	if(rq_init(dev))
	{
		fprintf(stderr,"rq_init failure\n");
		goto out;
	}

	if(kl_init(dev))
	{
		fprintf(stderr,"kl_init failure\n");
		goto out;
	}

	if(wl_init(dev))
	{
		fprintf(stderr,"wl_init failure\n");
		goto out;
	}

	if(out_init(script,cfn))
	{
		fprintf(stderr,"out_init failure\n");
		goto out;
	}

	err=0;
	goto done;

out:	mgr_exit();
	rq_exit();
	kl_exit();
	wl_exit();
	out_exit();

done:	return err;
}

static int ss(int mode)
{
	int err=-1;

	if(!mode)
	{
		err=0;
		goto out;
	}

	if(mgr_start())
	{
		fprintf(stderr,"mgr_start failure\n");
		goto err1;
	}

	if(out_start())
	{
		fprintf(stderr,"out_start failure\n");
		goto err2;
	}

	if(kl_start())
	{
		fprintf(stderr,"kl_start failure\n");
		goto err3;
	}

	if(rq_start())
	{
		fprintf(stderr,"rq_start failure\n");
		goto err4;
	}

	if(wl_start())
	{
		fprintf(stderr,"wl_start failure\n");
		goto err5;
	}

	err=0;
	goto err1;

out:	wl_stop();
err5:	rq_stop();
err4:	kl_stop();
err3:	out_stop();
err2:	mgr_stop();
err1:	return err;
}

static void usage(void)
{
	fprintf(stderr,"Usage: v6dnsd [-f] [-i device] [-B] [-c configfile] "
		"[-x script]\n"
	               "       v6dnsd -q [-i device]\n"
	               "       v6dnsd -4 ipv4-ns-address [...]\n"
	               "       v6dnsd -l\n"
	               "       v6dnsd -u\n"
	               "       v6dnsd -t\n"
	               "-f             don't daemonize, stay in foreground\n"
	               "-i device      wlan device (default: wlan0)\n"
	               "-B             set no burst mode as default\n"
	               "-c configfile  optional configuration file\n"
	               "-x script      script to execute (default: "
					"/etc/v6dnsd.sh)\n"
	               "-4             set wlan ipv4 name servers\n"
	               "-r             set radio name servers\n"
	               "-l             ouput current list of name servers\n"
	               "-u             reload optional configuration file\n"
	               "-t             terminate v6dnsd daemon\n");
}

int main(int argc,char *argv[])
{
	int c;
	int burst=1;
	int v4input=0;
	int radioinput=0;
	int nslist=0;
	int fg=0;
	int term=0;
	int update=0;
	sigset_t s;
	CFG *cfg=NULL;
	CFG *e;
	char *dev=NULL;
	char *x=NULL;
	char *cfn=NULL;

	while((c=getopt(argc,argv,"i:Bc:x:4rlftu"))!=-1)switch(c)
	{
	case 'i':
		if(dev)goto err;
		dev=optarg;
		break;

	case 'B':
		if(!burst)goto err;
		burst=0;
		break;

	case 'c':
		if(cfg)goto err;
		if(!(cfg=out_doparse(optarg)))goto fail;
		cfn=optarg;
		break;

	case 'x':
		if(x)goto err;
		x=optarg;
		break;

	case '4':
		if(v4input)goto err;
		v4input=1;
		break;

	case 'r':
		if(radioinput)goto err;
		radioinput=1;
		break;

	case 'l':
		if(nslist)goto err;
		nslist=1;
		break;

	case 'f':
		if(fg)goto err;
		fg=1;
		break;

	case 't':
		if(term)goto err;
		term=1;
		break;

	case 'u':
		if(update)goto err;
		update=1;
		break;

	default:printf("\n\n");
err:		usage();
		goto fail;
	}

	switch(v4input+radioinput+nslist+update+term)
	{
	case 1:	if(!(x||cfg||!burst||fg||dev))break;
	default:usage();
		goto fail;
	case 0:	break;
	}

	if(!dev)dev="wlan0";
	if(!x)x="/etc/v6dnsd.sh";

	if(optind>argc)optind=argc;
	argc-=optind;
	argv+=optind;

	if(v4input)exit(out_v4input(argc,argv));
	if(radioinput)exit(out_radioinput(argc,argv));
	if(nslist)exit(out_nslist());
	if(update)exit(out_newcfg());
	if(term)exit(out_doterm());

	if(access(x,X_OK))
	{
		usage();
		goto fail;
	}

	sigfillset(&s);
	sigprocmask(SIG_BLOCK,&s,NULL);

	signal(SIGINT,dummy);
	signal(SIGHUP,dummy);
	signal(SIGQUIT,dummy);
	signal(SIGTERM,dummy);

	if(infi(1,dev,burst,cfg,x,cfn))goto fail;

	if(!fg)daemon(0,0);

	if(ss(1))goto fail;

	sigfillset(&s);
	sigdelset(&s,SIGQUIT);
	sigdelset(&s,SIGTERM);
	sigdelset(&s,SIGHUP);
	sigdelset(&s,SIGINT);

	sigsuspend(&s);

	ss(0);

fail:	infi(0,NULL,0,NULL,NULL,NULL);
	while(cfg)
	{
		e=cfg;
		cfg=e->next;
		free(e);
	}
	exit(1);
}
