#include <pthread.h>
#include <sys/uio.h>
#include <poll.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>
#include "requester.h"
#include "manager.h"
#include "output.h"

#define DOWN	0
#define PROBE	1
#define ANSWER	2
#define INHIBIT	3
#define LIST	4
#define RENEW	5

#define MINWAIT	120
#define MAXWAIT	86400
#define RENTIME	60
#define INITIME	60

#define NOTX	-1
#define SHORTTX	3000
#define MEDTX	6000
#define LONGTX	600000

#define MEDTHR	10

typedef struct
{
	struct in6_addr	addr;
	int secs;
	int msecs;
	int used;
} ENTRY;

static int q[2]={-1,-1};
static timer_t tfd=-1;
static pthread_mutex_t mtx=PTHREAD_MUTEX_INITIALIZER;
static pthread_t wh;
static int armed=0;
static unsigned int expect=0;
static ENTRY list[MAX_NS];
static int state;
static int count;
static int isecs;
static int imsecs;
static int burst;
static int dftburst;
static NSLIST prev;
static CFG *cfg=NULL;

static void publish(int mode)
{
	int i;
	int j;
	V6NS v6;

	for(j=0,i=0;i<MAX_NS;i++)if(list[i].used)j++;

	if(j!=prev.total)mode=1;
	else
	{
		for(i=0;i<prev.total;i++)prev.valid[i]=0;
		for(i=0;i<MAX_NS;i++)if(list[i].used)for(j=0;j<prev.total;j++)
			if(!memcmp(&list[i].addr,&prev.addr[j],
				sizeof(struct in6_addr)))prev.valid[j]=1;
		for(j=0,i=0;i<prev.total;i++)if(prev.valid[i])j++;
		if(j!=i)mode=1;
	}

	if(mode)
	{
		for(prev.total=0,i=0;i<MAX_NS;i++)if(list[i].used)
			memcpy(&prev.addr[prev.total++],&list[i].addr,
				sizeof(struct in6_addr));

		for(v6.total=prev.total,i=0;i<v6.total;i++)
			memcpy(&v6.addr[i],&prev.addr[i],
				sizeof(struct in6_addr));

		out_cmd(OUT_V6LIST,&v6,sizeof(v6));
	}
}

static void expire(union sigval unused)
{
	unsigned int val;

	pthread_mutex_lock(&mtx);
	val=expect;
	pthread_mutex_unlock(&mtx);
	mgr_cmd(CMD_TIMER,&val,sizeof(unsigned int));
}

static void arm(int *secs,int *msecs)
{
	struct itimerspec t;
	struct itimerspec o;

	memset(&t,0,sizeof(t));
	if(timer_settime(tfd,0,&t,&o))
		perror("timer_settime");

	if(armed)
	{
		if(!o.it_value.tv_sec&&!o.it_value.tv_nsec)
		{
			pthread_mutex_lock(&mtx);
			expect++;
			pthread_mutex_unlock(&mtx);
		}
		armed=0;
	}

	t.it_value.tv_sec=*secs;
	t.it_value.tv_nsec=*msecs;

	*secs=o.it_value.tv_sec;
	*msecs=o.it_value.tv_nsec/1000000;
	if(imsecs<*msecs)
	{
		*secs=isecs-1-*secs;
		*msecs=imsecs+1000-*msecs;
	}
	else
	{
		*secs=isecs-*secs;
		*msecs=imsecs-*msecs;
	}

	isecs=t.it_value.tv_sec;
	imsecs=t.it_value.tv_nsec;

	if(!t.it_value.tv_sec&&!t.it_value.tv_nsec)return;

	t.it_value.tv_nsec*=1000000;
	armed=1;

	if(timer_settime(tfd,0,&t,NULL))
		perror("timer_settime");
}

static CFG *lookup(AP *a,CFG *list)
{
	for(;list;list=list->next)
	{
		if(!(list->flags&MGR_NOSSID))
			if(strcmp(a->ssid,list->ssid))continue;

		if(!(list->flags&MGR_NOBSSID))
			if(memcmp(a->ap,list->ap,BSSID_SIZE))continue;

		return list;
	}

	return NULL;
}

static void engine(int cmd,void *data)
{
	int i;
	int j;
	int secs;
	int msecs;
	AP *a;
	NSLIST *l;
	CFG *e;

	switch(cmd)
	{
	case CMD_AP:
		a=(AP *)data;

		switch(state)
		{
		case DOWN:
			if(a->valid)
			{
				e=lookup(a,cfg);
				if(e&&(e->flags&MGR_DISABLED))break;
				if(e)burst=e->burst;
				else burst=dftburst;

				state=PROBE;
				count=0;
				rq_delay(SHORTTX,burst);
				secs=INITIME;
				msecs=0;
				arm(&secs,&msecs);
			}
			break;

		default:
			e=lookup(a,cfg);
			if(e&&(e->flags&MGR_DISABLED))goto disable;

			if(!a->valid)
			{
disable:			rq_delay(NOTX,burst);
				secs=0;
				msecs=0;
				arm(&secs,&msecs);

				if(state==LIST||state==RENEW)
				{
					for(i=0;i<MAX_NS;i++)list[i].used=0;
					publish(0);
				}

				state=DOWN;
				break;
			}

			if(e)burst=e->burst;
			else burst=dftburst;

			if(state==LIST||state==RENEW)break;

			state=PROBE;
			count=0;
			rq_delay(SHORTTX,burst);
			secs=INITIME;
			msecs=0;
			arm(&secs,&msecs);
			break;
		}
		break;

	case CMD_TX:
		switch(state)
		{
		case PROBE:
		case RENEW:
			if(count<MEDTHR+1)count++;
			if(count==MEDTHR)rq_delay(MEDTX,burst);
			break;
		}
		break;

	case CMD_RX:
		switch(state)
		{
		case PROBE:
			rq_delay(LONGTX,burst);
			secs=0;
			msecs=0;
			arm(&secs,&msecs);
			state=ANSWER;
			break;
		}
		break;

	case CMD_NSLIST:
		l=(NSLIST*)data;

		switch(state)
		{
		default:
			rq_delay(NOTX,burst);
			secs=0;
			msecs=0;
			arm(&secs,&msecs);
			state=LIST;

			for(i=0;i<MAX_NS;i++)if(list[i].used)
			{
				if(secs>list[i].secs)list[i].used=0;
				else if(secs==list[i].secs&&
					msecs>=list[i].msecs)list[i].used=0;
				else
				{
					list[i].secs-=secs;
					if(msecs>list[i].msecs)
					{
						list[i].secs--;
						list[i].msecs+=1000;
					}
					list[i].msecs-=msecs;
				}
			}

			for(j=0;j<l->total;j++)for(i=0;i<MAX_NS;i++)
				if(list[i].used)if(!memcmp(&l->addr[j],
					&list[i].addr,sizeof(struct in6_addr)))
			{
				list[i].secs=l->valid[j];
				list[i].msecs=0;
				l->valid[j]=0;
				if(list[i].secs<MINWAIT)list[i].secs=MINWAIT;
				if(list[i].secs>MAXWAIT)list[i].secs=MAXWAIT;
				break;
			}

			for(j=0;j<l->total;j++)if(l->valid[j])
			{
				for(i=0;i<MAX_NS;i++)if(!list[i].used)break;
				if(i==MAX_NS)break;
				memcpy(&list[i].addr,&l->addr[j],
					sizeof(struct in6_addr));
				list[i].secs=l->valid[j];
				list[i].msecs=0;
				list[i].used=1;
				l->valid[j]=0;
				if(list[i].secs<MINWAIT)list[i].secs=MINWAIT;
				if(list[i].secs>MAXWAIT)list[i].secs=MAXWAIT;
			}

			for(j=0;j<l->total;j++)if(l->valid[j])
				for(i=0;i<MAX_NS;i++)
					if(l->valid[j]>list[i].secs)
			{
				memcpy(&list[i].addr,&l->addr[j],
					sizeof(struct in6_addr));
				list[i].secs=l->valid[j];
				list[i].msecs=0;
				if(list[i].secs<MINWAIT)list[i].secs=MINWAIT;
				if(list[i].secs>MAXWAIT)list[i].secs=MAXWAIT;
				break;
			}

			publish(0);

			secs=MAXWAIT;
			msecs=0;

			for(i=0;i<MAX_NS;i++)if(list[i].used)
			    if(list[i].secs<secs||
				(list[i].secs==secs&&list[i].msecs<msecs))
			{
				secs=list[i].secs;
				msecs=list[i].msecs;
			}

			if(secs<=RENTIME)
			{
				state=RENEW;
				count=0;
				rq_delay(SHORTTX,burst);
			}
			else secs-=RENTIME;

			arm(&secs,&msecs);


		case DOWN:
			break;
		}
		break;

	case CMD_TIMER:
		switch(state)
		{
		case PROBE:
			state=INHIBIT;
			rq_delay(NOTX,burst);
			secs=0;
			msecs=0;
			arm(&secs,&msecs);
			break;

		case LIST:
		case RENEW:
			rq_delay(NOTX,burst);
			secs=0;
			msecs=0;
			arm(&secs,&msecs);
			state=LIST;

			for(j=0,i=0;i<MAX_NS;i++)if(list[i].used)
			{
				if(secs>list[i].secs)
				{
					list[i].used=0;
					j=1;
				}
				else if(secs==list[i].secs&&
					msecs>=list[i].msecs)
				{
					list[i].used=0;
					j=1;
				}
				else
				{
					list[i].secs-=secs;
					if(msecs>list[i].msecs)
					{
						list[i].secs--;
						list[i].msecs+=1000;
					}
					list[i].msecs-=msecs;
				}
			}

			if(j)publish(0);

			secs=MAXWAIT;
			msecs=0;

			for(j=0,i=0;i<MAX_NS;i++)if(list[i].used)
			{
				if(list[i].secs<secs||
				    (list[i].secs==secs&&list[i].msecs<msecs))
				{	
					secs=list[i].secs;
					msecs=list[i].msecs;
				}
			}
			else j++;

			if(j==MAX_NS)
			{
				rq_delay(LONGTX,burst);
				state=ANSWER;
			}
			else
			{
				if(secs<=RENTIME)
				{
					state=RENEW;
					count=0;
					rq_delay(SHORTTX,burst);
				}
				else secs-=RENTIME;

				arm(&secs,&msecs);
			}
			break;
		}
		break;
	}
}

static void *worker(void *data)
{
	int secs;
	int msecs;
	unsigned int eid;
	int cmd;
	CFG *newcfg;
	CFG *e;
	struct pollfd p;
	sigset_t s;
	NSLIST l;
	AP a;

	sigfillset(&s);
	pthread_sigmask(SIG_BLOCK,&s,NULL);

	burst=dftburst;
	isecs=0;
	imsecs=0;
	state=DOWN;
	memset(&list,0,sizeof(list));
	memset(&prev,0,sizeof(prev));

	publish(1);

	p.fd=q[0];
	p.events=POLLIN;

	while(1)
	{
		if(poll(&p,1,-1)!=1)continue;

		if(!(p.revents&POLLIN))continue;

		if(read(q[0],&cmd,sizeof(int))!=sizeof(int))
		{
			perror("read");
			break;
		}

		switch(cmd)
		{
		case CMD_EXIT:
			goto out;

		case CMD_AP:
			if(read(q[0],&a,sizeof(AP))!=sizeof(AP))
			{
				perror("read");
				goto out;
			}
			engine(cmd,&a);
			break;

		case CMD_TX:
		case CMD_RX:
			engine(cmd,NULL);
			break;

		case CMD_NSLIST:
			if(read(q[0],&l,sizeof(NSLIST))!=sizeof(NSLIST))
			{
				perror("read");
				goto out;
			}
			engine(cmd,&l);
			break;

		case CMD_TIMER:
			if(read(q[0],&eid,sizeof(unsigned int))!=
				sizeof(unsigned int))
			{
				perror("read");
				goto out;
			}
			pthread_mutex_lock(&mtx);
			if(eid!=expect)
			{
				pthread_mutex_unlock(&mtx);
				break;
			}
			pthread_mutex_unlock(&mtx);
			engine(cmd,NULL);
			break;

		case CMD_NEWCFG:
			if(read(q[0],&newcfg,sizeof(newcfg))!=sizeof(newcfg))
			{
				perror("read");
				goto out;
			}

			while(cfg)
			{
				e=cfg;
				cfg=e->next;
				free(e);
			}

			cfg=newcfg;
			break;
		}
	}

out:	secs=0;
	msecs=0;
	arm(&secs,&msecs);
	rq_delay(NOTX,burst);

	pthread_exit(NULL);
	return NULL;
}

void mgr_exit(void)
{
	if(q[0]!=-1)close(q[0]);
	if(q[1]!=-1)close(q[1]);
	if(tfd!=-1)timer_delete(tfd);
	q[0]=-1;
	q[1]=-1;
	tfd=-1;
	armed=0;
	expect=0;
	cfg=NULL;
}

int mgr_init(int burst,CFG *cfg)
{
	struct sigevent s;

	dftburst=burst;

	if(q[0]==-1)
	{
		if(pipe2(q,O_CLOEXEC))
		{
			perror("pipe2");
			mgr_exit();
			return -1;
		}
	}

	if(tfd==-1)
	{
		memset(&s,0,sizeof(s));
		s.sigev_notify=SIGEV_THREAD;
		s.sigev_notify_function=expire;
		if(timer_create(CLOCK_MONOTONIC,&s,&tfd))
		{
			perror("timer_create");
			mgr_exit();
			return -1;
		}
	}

	return 0;
}

int mgr_start(void)
{
	if(pthread_create(&wh,NULL,worker,NULL))
	{
		perror("pthread_create");
		return -1;
	}

	return 0;
}

void mgr_stop(void)
{
	mgr_cmd(CMD_EXIT,NULL,0);
	pthread_join(wh,NULL);
}

void mgr_cmd(int cmd,void *data,int len)
{
	struct iovec iov[2];

	iov[0].iov_base=&cmd;
	iov[0].iov_len=sizeof(int);
	iov[1].iov_base=data;
	iov[1].iov_len=len;

	pthread_mutex_lock(&mtx);

	if(writev(q[1],iov,len?2:1)!=len+sizeof(int))
		perror("writev");

	pthread_mutex_unlock(&mtx);
}
