#include <pthread.h>
#include <sys/uio.h>
#include <poll.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>
#include "manager.h"
#include "output.h"
#include "missing.h"

#define SOCK "/dev/v6ctl"

static int sfd=-1;
static int q[2]={-1,-1};
static char *script=NULL;
static char *cfg=NULL;
static V4NS v4list;
static V6NS v6list;
static V4NS r4list;
static V6NS r6list;
static pthread_t wh;
static pthread_mutex_t mtx=PTHREAD_MUTEX_INITIALIZER;
static char b4[256];
static char b6[256];
static char r4[256];
static char r6[256];
static char bfr[1024];

static int clisend(int cmd,void *data,int len)
{
	int i;
	struct sockaddr_un a;
	struct iovec iov[2];

	memset(&a,0,sizeof(a));
	a.sun_family=AF_UNIX;
	strcpy(a.sun_path,SOCK);

	iov[0].iov_base=&cmd;
	iov[0].iov_len=sizeof(int);
	iov[1].iov_base=data;
	iov[1].iov_len=len;

	if((i=socket(PF_UNIX,SOCK_STREAM|SOCK_CLOEXEC,0))==-1)
	{
		perror("socket");
		return -1;
	}

	if(connect(i,(struct sockaddr *)(&a),sizeof(a)))
	{
		perror("connect");
		close(i);
		return -11;
	}

	if(writev(i,iov,data?2:1)!=(data?len:0)+sizeof(int))
	{
		perror("writev");
		close(i);
		return -1;
	}

	return i;
}

static int h2b(char *in,unsigned char *out)
{
	if(*in>='0'&&*in<='9')*out=*in++-'0';
	else if(*in>='A'&&*in<='F')*out=*in-'A'+10;
	else if(*in>='a'&&*in<='f')*out=*in-'a'+10;
	else return -1;

	*out<<=4;

	if(*in>='0'&&*in<='9')*out|=*in++-'0';
	else if(*in>='A'&&*in<='F')*out|=*in-'A'+10;
	else if(*in>='a'&&*in<='f')*out|=*in-'a'+10;
	else return -1;

	return 0;
}

static CFG *parse(char *cfg)
{
	int err=-1;
	int i;
	int line;
	int dftburst;
	int flags;
	FILE *fp;
	char *ssid;
	char *bssid;
	char *burst;
	CFG *c=NULL;
	CFG *e;
	CFG **a=&c;
	unsigned char addr[BSSID_SIZE];

	if(!(fp=fopen(cfg,"r")))
	{
		fprintf(stderr,"can't access %s\n",cfg);
		goto fail;
	}

	for(line=1,flags=0;fgets(bfr,sizeof(bfr),fp);line++,flags=0)
	{
		ssid=strtok(bfr," \t\r\n");
		if(!ssid||!*ssid||*ssid=='#')continue;

		if(!strcmp(ssid,"+"))
		{
			flags|=MGR_NOSSID;
			*ssid=0;
		}

		if(strlen(ssid)>SSID_MAX)
		{
			fprintf(stderr,"illegal ssid line %d of %s\n",line,cfg);
			goto fail;
		}

		bssid=strtok(NULL," \t\r\n");
		burst=strtok(NULL," \t\r\n");

		if(!bssid||!*bssid||!burst||!*burst)
		{
			fprintf(stderr,"syntax error line %d of %s\n",line,cfg);
			goto fail;
		}

		if(!strcmp(bssid,"+"))
		{
			flags|=MGR_NOBSSID;
			memset(addr,0,sizeof(addr));
		}
		else for(i=0;i<BSSID_SIZE;i++)
		{
			if(h2b(bssid,addr+i))goto bsserr;
			bssid+=2;

			if(i==5)
			{
				if(*bssid)goto bsserr;
			}
			else if(*bssid++!=':')
			{
bsserr:				fprintf(stderr,"illegal bssid line %d of %s\n",
					line,cfg);
				goto fail;
			}
		}

		if(!strcmp(burst,"1"))dftburst=1;
		else if(!strcmp(burst,"0"))dftburst=0;
		else if(!strcmp(burst,"-1"))
		{
			flags|=MGR_DISABLED;
			dftburst=-1;
		}
		else
		{
			fprintf(stderr,"illegal burst line %d of %s\n",
				line,cfg);
			goto fail;
		}

		if(!(e=malloc(sizeof(CFG))))
		{
			perror("malloc");
			goto fail;
		}

		e->next=NULL;
		e->flags=flags;
		e->burst=dftburst;
		memcpy(e->ap,addr,sizeof(addr));
		strcpy(e->ssid,ssid);

		*a=e;
		a=&e->next;

		if((flags&(MGR_NOSSID|MGR_NOBSSID))==(MGR_NOSSID|MGR_NOBSSID))
			break;
	}

	err=0;

fail:	if(err)while(c)
	{
		e=c;
		c=e->next;
		free(e);
	}

	fclose(fp);

	return c;
}

static int v4cmp(V4NS *l1,V4NS *l2)
{
	int i;
	int j;
	int hit[MAX_NS];

	if(l1->total!=l2->total)return 1;

	memset(hit,0,sizeof(hit));

	for(i=0;i<l1->total;i++)for(j=0;j<l2->total;j++)if(!hit[j])
		if(l1->addr[i].s_addr==l2->addr[j].s_addr)
	{
		hit[j]=1;
		break;
	}

	for(i=0;i<l2->total;i++)if(!hit[i])return 1;

	return 0;
}

static int v6cmp(V6NS *l1,V6NS *l2)
{
	int i;
	int j;
	int hit[MAX_NS];

	if(l1->total!=l2->total)return 1;

	memset(hit,0,sizeof(hit));

	for(i=0;i<l1->total;i++)for(j=0;j<l2->total;j++)if(!hit[j])
		if(!memcmp(l1->addr[i].s6_addr,l2->addr[j].s6_addr,
			sizeof(struct in6_addr)))
	{
		hit[j]=1;
		break;
	}

	for(i=0;i<l2->total;i++)if(!hit[i])return 1;

	return 0;
}

static void *worker(void *data)
{
	int c;
	int i;
	int cmd;
	int update;
	int status;
	pid_t child;
	CFG *newcfg;
	struct pollfd p[2];
	struct iovec iov[2];
	V6NS v6;
	V4NS v4;
	V4NS prev;
	V4NS rprv;
	V6NS rprv6;
	sigset_t s;
	char bfr[INET6_ADDRSTRLEN];

	sigfillset(&s);
	pthread_sigmask(SIG_BLOCK,&s,NULL);

	memset(&prev,0,sizeof(prev));
	memset(&rprv,0,sizeof(rprv));
	memset(&rprv6,0,sizeof(rprv6));
	memset(&v4list,0,sizeof(v4list));
	memset(&v6list,0,sizeof(v6list));
	memset(&r4list,0,sizeof(r4list));
	memset(&r6list,0,sizeof(r6list));

	p[0].fd=sfd;
	p[0].events=POLLIN;
	p[1].fd=q[0];
	p[1].events=POLLIN;

	update=0;
	child=0;
	while(1)
	{
		if(child)if(waitpid(child,&status,WNOHANG)==child)child=0;

		if(update&&!child)switch((child=fork()))
		{
		case -1:perror("fork");
			child=0;
			break;

		case 0:	strcpy(b6,"V6DNSD_V6=");
			for(i=0;i<v6list.total;i++)
			{
				inet_ntop(AF_INET6,v6list.addr[i].s6_addr,bfr,
					INET6_ADDRSTRLEN);
				if(i)strcat(b6," ");
				strcat(b6,bfr);
			}
			if(putenv(b6))
			{
				perror("putenv");
				_exit(255);
			}

			strcpy(b4,"V6DNSD_V4=");
			for(i=0;i<v4list.total;i++)
			{
				inet_ntop(AF_INET,&v4list.addr[i].s_addr,bfr,
					INET_ADDRSTRLEN);
				if(i)strcat(b4," ");
				strcat(b4,bfr);
			}
			if(putenv(b4))
			{
				perror("putenv");
				_exit(255);
			}

			strcpy(r6,"V6DNSD_R6=");
			for(i=0;i<r6list.total;i++)
			{
				inet_ntop(AF_INET6,r6list.addr[i].s6_addr,bfr,
					INET6_ADDRSTRLEN);
				if(i)strcat(r6," ");
				strcat(r6,bfr);
			}
			if(putenv(r6))
			{
				perror("putenv");
				_exit(255);
			}

			strcpy(r4,"V6DNSD_R4=");
			for(i=0;i<r4list.total;i++)
			{
				inet_ntop(AF_INET,&r4list.addr[i].s_addr,bfr,
					INET_ADDRSTRLEN);
				if(i)strcat(r4," ");
				strcat(r4,bfr);
			}
			if(putenv(r4))
			{
				perror("putenv");
				_exit(255);
			}

			execl(script,script,NULL);
			perror("execl");
			_exit(255);

		default:update=0;
			break;
		}

		if(poll(p,2,update||child?250:-1)!=1)continue;

		if(p[0].revents&POLLIN)if((c=accept(sfd,NULL,NULL))!=-1)
		{
			if(fcntl(c,F_SETFD,FD_CLOEXEC))
			{
				perror("fcntl");
				goto done;
			}

			cmd=0;
			if(ioctl(c,FIONBIO,&cmd))
			{
				perror("ioctl");
				goto done;
			}

			if(read(c,&cmd,sizeof(int))!=sizeof(int))goto done;

			switch(cmd)
			{
			case OUT_V4LIST:
				if(read(c,&v4,sizeof(v4))!=sizeof(v4))
				{
					perror("read");
					goto done;
				}

				if(v4cmp(&v4,&prev))
				{
					memcpy(&prev,&v4,sizeof(v4));
					memcpy(&v4list,&v4,sizeof(v4));
					update=1;
				}
				break;

			case OUT_RADIOLIST:
				if(read(c,&v4,sizeof(v4))!=sizeof(v4))
				{
					perror("read");
					goto done;
				}

				if(read(c,&v6,sizeof(v6))!=sizeof(v6))
				{
					perror("read");
					goto done;
				}

				if(v4cmp(&v4,&rprv))
				{
					memcpy(&rprv,&v4,sizeof(v4));
					memcpy(&r4list,&v4,sizeof(v4));
					update=1;
				}

				if(v6cmp(&v6,&rprv6))
				{
					memcpy(&rprv6,&v6,sizeof(v6));
					memcpy(&r6list,&v6,sizeof(v6));
					update=1;
				}
				break;

			case OUT_NSLIST:
				iov[0].iov_base=&v6list;
				iov[0].iov_len=sizeof(v6list);
				iov[1].iov_base=&v4list;
				iov[1].iov_len=sizeof(v4list);

				if(writev(c,iov,2)!=sizeof(v6list)+
					sizeof(v4list))perror("writev");
				break;

			case OUT_TERM:
				kill(getpid(),SIGTERM);
				break;

			case OUT_RELOAD:
				if(cfg)if((newcfg=parse(cfg)))
					mgr_cmd(CMD_NEWCFG,&newcfg,
						sizeof(newcfg));
				break;
			}

done:			close(c);
		}

		if(!(p[1].revents&POLLIN))continue;

		if(read(q[0],&cmd,sizeof(int))!=sizeof(int))
		{
			perror("read");
			break;
		}

		switch(cmd)
		{
		case OUT_EXIT:
			goto out;

		case OUT_V6LIST:
			if(read(q[0],&v6,sizeof(v6))!=sizeof(v6))
			{
				perror("read");
				goto out;
			}

			memcpy(&v6list,&v6,sizeof(v6));
			update=1;
			break;

		}
	}

out:	pthread_exit(NULL);
	return NULL;
}

void out_exit(void)
{
	if(q[0]!=-1)close(q[0]);
	if(q[1]!=-1)close(q[1]);
	if(sfd!=-1)
	{
		close(sfd);
		unlink(SOCK);
	}
	if(script)free(script);
	if(cfg)free(cfg);
	q[0]=-1;
	q[1]=-1;
	sfd=-1;
	script=NULL;
	cfg=NULL;
}

int out_init(char *scr,char *cfn)
{
	int c;
	struct sockaddr_un a;
	mode_t m;

	if(script)free(script);
	if(!(script=strdup(scr)))
	{
		perror("strdup");
		out_exit();
		return -1;
	}

	if(cfg)free(cfg);
	if(!cfn)cfg=NULL;
	else if(!(cfg=strdup(cfn)))
	{
		perror("strdup");
		out_exit();
		return -1;
	}

	if(q[0]==-1)
	{
		if(pipe2(q,O_CLOEXEC))
		{
			perror("pipe2");
			out_exit();
			return -1;
		}
	}

	if(sfd==-1)
	{
		memset(&a,0,sizeof(a));
		a.sun_family=AF_UNIX;
		strcpy(a.sun_path,SOCK);

		if(!access(SOCK,W_OK))
		{
			if((c=socket(PF_UNIX,SOCK_STREAM|SOCK_CLOEXEC,0))==-1)
			{
				perror("socket");
				out_exit();
				return -1;
			}

			if(!connect(c,(struct sockaddr *)(&a),sizeof(a)))
			{
				close(c);
				fprintf(stderr,"already running\n");
				out_exit();
				return -1;
			}

			close(c);
			unlink(SOCK);
		}

		if((sfd=socket(PF_UNIX,SOCK_STREAM|SOCK_CLOEXEC|SOCK_NONBLOCK,
			0))==-1)
		{
			perror("socket");
			out_exit();
			return -1;
		}

		m=umask(0);
		if(bind(sfd,(struct sockaddr *)(&a),sizeof(a)))
		{
			perror("bind");
			umask(m);
			out_exit();
			return -1;
		}
		umask(m);

		if(listen(sfd,5))
		{
			perror("listen");
			out_exit();
			return -1;
		}
	}

	return 0;
}

int out_start(void)
{
	if(pthread_create(&wh,NULL,worker,NULL))
	{
		perror("pthread_create");
		return -1;
	}

	return 0;
}

void out_stop(void)
{
	out_cmd(OUT_EXIT,NULL,0);
	pthread_join(wh,NULL);
}

void out_cmd(int cmd,void *data,int len)
{
	struct iovec iov[2];

	iov[0].iov_base=&cmd;
	iov[0].iov_len=sizeof(int);
	iov[1].iov_base=data;
	iov[1].iov_len=len;

	pthread_mutex_lock(&mtx);

	if(writev(q[1],iov,len?2:1)!=len+sizeof(int))
		perror("writev");

	pthread_mutex_unlock(&mtx);
}

int out_v4input(int argc,char *argv[])
{
	int i;
	V4NS v4;

	for(i=0;i<argc&&i<4;i++)if(inet_pton(AF_INET,argv[i],&v4.addr[i])!=1)
	{
		fprintf(stderr,"illegal address %s\n",argv[i]);
		return 1;
	}
	v4.total=i;

	if((i=clisend(OUT_V4LIST,&v4,sizeof(v4)))==-1)return 1;

	close(i);

	return 0;
}

int out_radioinput(int argc,char *argv[])
{
	int i;
	struct in_addr a4;
	struct in6_addr a6;
	struct
	{
		V4NS v4;
		V6NS v6;
	} data;

	for(data.v4.total=0,data.v6.total=0,i=0;i<argc;i++)
	{
		if(inet_pton(AF_INET,argv[i],&a4)==1)
		{
			if(data.v4.total<4)data.v4.addr[data.v4.total++]=a4;
		}
		else if(inet_pton(AF_INET6,argv[i],&a6)==1)
		{
			if(data.v6.total<4)data.v6.addr[data.v6.total++]=a6;
		}
		else
		{
			fprintf(stderr,"illegal address %s\n",argv[i]);
			return 1;
		}
	}

	if((i=clisend(OUT_RADIOLIST,&data,sizeof(data)))==-1)return 1;

	close(i);

	return 0;
}

int out_nslist(void)
{
	int i;
	V4NS v4;
	V6NS v6;
	char bfr[INET6_ADDRSTRLEN];

	if((i=clisend(OUT_NSLIST,NULL,0))==-1)return 1;

	if(read(i,&v6,sizeof(v6))!=sizeof(v6))
	{
		perror("read");
		close(i);
		return 1;
	}

	if(read(i,&v4,sizeof(v4))!=sizeof(v4))
	{
		perror("read");
		close(i);
		return 1;
	}

	close(i);

	for(i=0;i<v6.total;i++)
	{
		inet_ntop(AF_INET6,v6.addr[i].s6_addr,bfr,INET6_ADDRSTRLEN);
		printf("%s\n",bfr);
	}

	for(i=0;i<v4.total;i++)
	{
		inet_ntop(AF_INET,&v4.addr[i].s_addr,bfr,INET_ADDRSTRLEN);
		printf("%s\n",bfr);
	}

	return 0;
}

int out_doterm(void)
{
	int i;

	if((i=clisend(OUT_TERM,NULL,0))==-1)return 1;

	close(i);

	return 0;
}

int out_newcfg(void)
{
	int i;

	if((i=clisend(OUT_RELOAD,NULL,0))==-1)return 1;

	close(i);

	return 0;
}

CFG *out_doparse(char *cfg)
{
	return parse(cfg);
}
