#include <pthread.h>
#include <poll.h>
#include <errno.h>
#include <stdio.h>
#include <signal.h>
#include <sys/eventfd.h>
#include <netdb.h>
#include <net/if.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/icmp6.h>
#include "requester.h"
#include "manager.h"
#include "missing.h"

#define MINOK	3000
#define MINERR	6000

static int rxpkt;
static int dorep=0;
static int delay=-1;
static int tfd=-1;
static int rfd=-1;
static int sfd=-1;
static char *dev=NULL;
static pthread_mutex_t mtx=PTHREAD_MUTEX_INITIALIZER;
static pthread_t th;
static pthread_t rh;
static unsigned char bfr[16384];

static int getsrcaddr(char *ifc,struct sockaddr_in6 *a)
{
	struct addrinfo h;
	struct addrinfo *r;

	memset(&h,0,sizeof(struct addrinfo));
	h.ai_family=AF_INET6;
	h.ai_socktype=SOCK_DGRAM;
	h.ai_flags=AI_NUMERICHOST;

	if(getaddrinfo("ff02::2",NULL,&h,&r))
	{
		perror("getaddrinfo");
		return -1;
	}
	memcpy(a,r->ai_addr,sizeof(struct sockaddr_in6));
	freeaddrinfo(r);

	if(!(a->sin6_scope_id=if_nametoindex(ifc)))
	{
		perror("if_nametoindex");
		return -1;
	}

	return 0;
}

static void *rxworker(void *data)
{
	int len;
	struct pollfd p[2];
	uint64_t dummy;
	sigset_t s;
	struct iovec iov;
	struct msghdr hdr;
	struct cmsghdr *cmsg;
	struct nd_router_advert *ra;
	struct sockaddr_in6 addr;
	char cbfr[CMSG_SPACE(sizeof (int))];

	sigfillset(&s);
	pthread_sigmask(SIG_BLOCK,&s,NULL);

	p[0].fd=rfd;
	p[0].events=POLLIN;
	p[1].fd=sfd;
	p[1].events=POLLIN;

	while(1)
	{
repeat:		if(poll(p,2,-1)<=0)continue;

		if(p[0].revents&POLLIN)
		{
			if(read(rfd,&dummy,sizeof(dummy))!=sizeof(dummy))
				perror("read eventfd");
			break;
		}

		if(p[1].revents&POLLIN)
		{
			iov.iov_base=bfr;
			iov.iov_len=sizeof(bfr);

			hdr.msg_name=&addr;
			hdr.msg_namelen=sizeof(addr);
			hdr.msg_iov=&iov;
			hdr.msg_iovlen=1;
			hdr.msg_control=&cbfr;
			hdr.msg_controllen=sizeof(cbfr);

			if((len=recvmsg(sfd,&hdr,MSG_DONTWAIT))==-1)continue;

			for(cmsg=CMSG_FIRSTHDR(&hdr);cmsg;
				cmsg=CMSG_NXTHDR(&hdr,cmsg))
				if(cmsg->cmsg_level==IPPROTO_IPV6
					&&cmsg->cmsg_type==IPV6_HOPLIMIT)
						if(*(int *)CMSG_DATA(cmsg)!=255)
							goto repeat;

			ra=(struct nd_router_advert *)bfr;

			if(len<sizeof(struct nd_router_advert)||
				ra->nd_ra_type!=ND_ROUTER_ADVERT||
				ra->nd_ra_code)continue;

			if(addr.sin6_scope_id==if_nametoindex(dev))
			{
				pthread_mutex_lock(&mtx);
				rxpkt=1;
				pthread_mutex_unlock(&mtx);
				mgr_cmd(CMD_RX,NULL,0);
			}
		}
	}

	pthread_exit(NULL);
	return NULL;
}

static void *txworker(void *data)
{
	int repeat;
	struct pollfd p;
	int dly;
	int min;
	uint64_t dummy;
	sigset_t s;
	struct sockaddr_in6 a;
	struct nd_router_solicit pkt;

	sigfillset(&s);
	pthread_sigmask(SIG_BLOCK,&s,NULL);

	min=MINOK;

	p.fd=tfd;
	p.events=POLLIN;

	while(1)
	{
		pthread_mutex_lock(&mtx);
		dly=delay;
		pthread_mutex_unlock(&mtx);

		if(dly!=-1)if(dly<min)dly=min;

		if(poll(&p,1,dly)==1)if(p.revents&POLLIN)
		{
			if(read(tfd,&dummy,sizeof(dummy))!=sizeof(dummy))
			{
				perror("read eventfd");
				break;
			}
			min=MINOK;
			if(dummy<256)continue;
			break;
		}

		min=MINERR;

		if(getsrcaddr(dev,&a))continue;

		pthread_mutex_lock(&mtx);
		repeat=dorep;
		rxpkt=0;
		pthread_mutex_unlock(&mtx);

		memset(&pkt,0,sizeof(pkt));
        	pkt.nd_rs_type=ND_ROUTER_SOLICIT;

again:		if(sendto(sfd,&pkt,sizeof(pkt),0,(struct sockaddr *)&a,
			sizeof(a))!=sizeof(pkt))
		{
			perror("sendto ND_ROUTER_SOLICIT");
			continue;
		}

		if(repeat--)
		{
			usleep(20000);
			pthread_mutex_lock(&mtx);
			if(!rxpkt)
			{
				pthread_mutex_unlock(&mtx);
				goto again;
			}
			pthread_mutex_unlock(&mtx);
		}

		min=MINOK;

		mgr_cmd(CMD_TX,NULL,0);
	}

	pthread_exit(NULL);
	return NULL;
}

void rq_exit(void)
{
	if(tfd!=-1)close(tfd);
	if(rfd!=-1)close(rfd);
	if(sfd!=-1)close(sfd);
	if(dev)free(dev);
	tfd=-1;
	rfd=-1;
	sfd=-1;
	delay=-1;
	dorep=0;
	dev=NULL;
}

int rq_init(char *ifc)
{
	int i;

	if(dev)free(dev);
	dev=strdup(ifc);
	if(!dev)
	{
		perror("strdup");
		rq_exit();
		return -1;
	}

	if(tfd==-1)
	{
		tfd=eventfd(0,EFD_CLOEXEC);
		if(tfd==-1)
		{
			perror("eventfd");
			rq_exit();
			return -1;
		}
	}

	if(rfd==-1)
	{
		rfd=eventfd(0,EFD_CLOEXEC);
		if(rfd==-1)
		{
			perror("eventfd");
			rq_exit();
			return -1;
		}
	}

	if(sfd==-1)
	{
		sfd=socket(PF_INET6,SOCK_RAW|SOCK_CLOEXEC,IPPROTO_ICMPV6);
		if(sfd==-1)
		{
			perror("socket SOCK_RAW");
			rq_exit();
			return -1;
		}

		i=1;
		if(setsockopt(sfd,SOL_SOCKET,SO_DONTROUTE,&i,sizeof(i)))
		{
			perror("setsockopt SO_DONTROUTE");
			rq_exit();
			return -1;
		}

		i=255;
		if(setsockopt(sfd,IPPROTO_IPV6,IPV6_MULTICAST_HOPS,&i,
			sizeof(i)))
		{
			perror("setsockopt IPV6_MULTICAST_HOPS");
			rq_exit();
			return -1;
		}
		if(setsockopt(sfd,IPPROTO_IPV6,IPV6_UNICAST_HOPS,&i,sizeof(i)))
		{
			perror("setsockopt IPV6_UNICAST_HOPS");
			rq_exit();
			return -1;
		}

		i=1;
		if(setsockopt(sfd,IPPROTO_IPV6,IPV6_RECVHOPLIMIT,&i,sizeof(i)))
		{
			perror("setsockopt IPV6_RECVHOPLIMIT");
			rq_exit();
			return -1;
		}

	}

	return 0;
}

int rq_start(void)
{
	uint64_t dummy=256;

	if(pthread_create(&th,NULL,txworker,NULL))
	{
		perror("pthread_create");
		return -1;
	}

	if(pthread_create(&rh,NULL,rxworker,NULL))
	{
		perror("pthread_create");
		pthread_mutex_lock(&mtx);
		write(tfd,&dummy,sizeof(dummy));
		pthread_mutex_unlock(&mtx);
		pthread_join(th,NULL);
		return -1;
	}

	return 0;
}

void rq_stop(void)
{
	uint64_t dummy=256;

	pthread_mutex_lock(&mtx);
	write(tfd,&dummy,sizeof(dummy));
	write(rfd,&dummy,sizeof(dummy));
	pthread_mutex_unlock(&mtx);
	pthread_join(th,NULL);
	pthread_join(rh,NULL);
}

void rq_delay(int dly,int burst)
{
	uint64_t dummy=1;

	pthread_mutex_lock(&mtx);
	delay=dly;
	dorep=burst?25:0;
	pthread_mutex_unlock(&mtx);
	write(tfd,&dummy,sizeof(dummy));
}
