#include <pthread.h>
#include <unistd.h>
#include <signal.h>
#include <poll.h>
#include <errno.h>
#include <stdio.h>
#include <sys/socket.h>
#include <linux/rtnetlink.h>
#include <netinet/in.h>
#include <netinet/icmp6.h>
#include <net/if.h>
#include <netinet/ether.h>
#include <sys/eventfd.h>
#include "wireless.h"
#include "iwlib.h"
#include "manager.h"
#include "missing.h"

static int efd=-1;
static int sfd=-1;
static int vers=-1;
static int opmode=-1;
static char *dev=NULL;
static pthread_t wh;
static AP addr;
static char bfr[8192];

static int getssid(char *ifc)
{
	int s;
	struct iwreq wrq;

	if((s=socket(AF_INET6,SOCK_DGRAM,0))==-1)
	{
		perror("socket");
		return -1;
	}

	memset(addr.ssid,0,sizeof(addr.ssid));
	memset(&wrq,0,sizeof(wrq));
	wrq.u.data.pointer=addr.ssid;
	wrq.u.data.length=sizeof(addr.ssid);
	strncpy(wrq.ifr_name,ifc,IFNAMSIZ);

	if(ioctl(s,SIOCGIWESSID,&wrq))
	{
		perror("ioctl SIOCGIWESSID");
		close(s);
		return -1;
	}

	close(s);
	return 0;
}

static int apvalid(void)
{
	int i;

	for(i=0;i<BSSID_SIZE;i++)if(addr.ap[i])break;
	if(i==BSSID_SIZE)return 0;
	for(i=0;i<BSSID_SIZE;i++)if(addr.ap[i]!=0x44)break;
	if(i==BSSID_SIZE)return 0;
	for(i=0;i<BSSID_SIZE;i++)if(addr.ap[i]!=0xff)break;
	if(i==BSSID_SIZE)return 0;
	return 1;
}

static void *worker(void *data)
{
	struct pollfd p[2];
	int len;
	uint64_t dummy;
	sigset_t s;
	struct nlmsghdr *h;
	struct ifinfomsg* ifi;
	int attrlen;
	struct rtattr *attr;
	struct stream_descr sd;
	struct iw_event iw;
	struct ether_addr *eth;

	sigfillset(&s);
	pthread_sigmask(SIG_BLOCK,&s,NULL);

	p[0].fd=efd;
	p[0].events=POLLIN;
	p[1].fd=sfd;
	p[1].events=POLLIN;

	mgr_cmd(CMD_AP,&addr,sizeof(AP));

	while(1)
	{
		if(poll(p,2,-1)<=0)continue;

		if(p[0].revents&POLLIN)
		{
			if(read(efd,&dummy,sizeof(dummy))!=sizeof(dummy))
				perror("read eventfd");
			break;
		}

		if(p[1].revents&POLLIN)
		{
			if((len=recv(sfd,bfr,sizeof(bfr),0))==-1)
			{
				perror("recv");
				break;
			}

			for(h=(struct nlmsghdr *)bfr;len>=sizeof(*h);)
			{
				if(h->nlmsg_type!=RTM_NEWLINK)goto next;

				if(h->nlmsg_len<=
					NLMSG_ALIGN(sizeof(struct ifinfomsg)))
						goto next;

				ifi=NLMSG_DATA(h);
				if(ifi->ifi_index!=if_nametoindex(dev))
					goto next;

				attrlen=h->nlmsg_len-
					NLMSG_ALIGN(sizeof(struct ifinfomsg));
				attr=(void *)((char *)ifi+
					NLMSG_ALIGN(sizeof(struct ifinfomsg)));

				for(;RTA_OK(attr,attrlen);
				    attr=RTA_NEXT(attr,attrlen))
					if(attr->rta_type==IFLA_WIRELESS)
				{
					iw_init_event_stream(&sd,(char *)attr+
					    RTA_ALIGN(sizeof(struct rtattr)),
					    attr->rta_len-
					    RTA_ALIGN(sizeof(struct rtattr)));

					while(iw_extract_event_stream(&sd,&iw,
						vers)>0)switch(iw.cmd)
					{
					case SIOCSIWMODE:
						opmode=iw.u.mode;
						break;

					case SIOCGIWAP:
						eth=(struct ether_addr *)
						    &iw.u.ap_addr.sa_data;
						memcpy(addr.ap,
						    eth->ether_addr_octet,
						    BSSID_SIZE);
						addr.valid=apvalid();

						if(addr.valid)getssid(dev);

						if(opmode!=IW_MODE_INFRA)
							addr.valid=0;

						mgr_cmd(CMD_AP,&addr,
							sizeof(AP));
						break;
					}
				}

next:				len-=NLMSG_ALIGN(h->nlmsg_len);
				h=(struct nlmsghdr *)
					((char *)h+NLMSG_ALIGN(h->nlmsg_len));
			}
		}
	}

	pthread_exit(NULL);
	return NULL;
}

static int checkvers(void)
{
	int s;
	struct iwreq wrq;
	struct iw_range *rr;
	struct ether_addr *eth;

	if((s=socket(AF_INET6,SOCK_DGRAM,0))==-1)
	{
		perror("socket");
		return -1;
	}

	memset(&wrq,0,sizeof(wrq));
	wrq.u.data.pointer=bfr;
	wrq.u.data.length=sizeof(bfr);
	strncpy(wrq.ifr_name,dev,IFNAMSIZ);

	if(ioctl(s,SIOCGIWRANGE,&wrq))
	{
		perror("ioctl SIOCGIWRANGE");
		close(s);
		return -1;
	}

	rr=(struct iw_range *)bfr;

	if(wrq.u.data.length<300)vers=9;
	else vers=rr->we_version_compiled;

	memset(&wrq,0,sizeof(wrq));
	wrq.u.data.pointer=bfr;
	wrq.u.data.length=sizeof(bfr);
	strncpy(wrq.ifr_name,dev,IFNAMSIZ);

	if(ioctl(s,SIOCGIWAP,&wrq))
	{
		perror("ioctl SIOCGIWAP");
		close(s);
		return -1;
	}

	eth=(struct ether_addr *)&wrq.u.ap_addr.sa_data;
	memcpy(addr.ap,eth->ether_addr_octet,BSSID_SIZE);
	addr.valid=apvalid();

	memset(&wrq,0,sizeof(wrq));
	wrq.u.data.pointer=bfr;
	wrq.u.data.length=sizeof(bfr);
	strncpy(wrq.ifr_name,dev,IFNAMSIZ);

	if(ioctl(s,SIOCGIWMODE,&wrq))
	{
		perror("ioctl SIOCGIWMODE");
		close(s);
		return -1;
	}

	opmode=wrq.u.mode;

	close(s);

	if(addr.valid)if(getssid(dev))return -1;

	if(vers<19)
	{
		fprintf(stderr,"kernel wireless version too old (%d)\n",vers);
		return -1;
	}

	if(opmode!=IW_MODE_INFRA)addr.valid=0;

	return 0;
}

void wl_exit(void)
{
	if(efd!=-1)close(efd);
	if(sfd!=-1)close(sfd);
	if(dev)free(dev);
	efd=-1;
	sfd=-1;
	opmode=-1;
	dev=NULL;
}

int wl_init(char *ifc)
{
	struct sockaddr_nl a;

	if(dev)free(dev);
	dev=strdup(ifc);
	if(!dev)
	{
		perror("strdup");
		wl_exit();
		return -1;
	}

	if(efd==-1)
	{
		efd=eventfd(0,EFD_CLOEXEC);
		if(efd==-1)
		{
			perror("eventfd");
			wl_exit();
			return -1;
		}
	}

	if(sfd==-1)
	{
		sfd=socket(PF_NETLINK,SOCK_RAW|SOCK_CLOEXEC,NETLINK_ROUTE);
		if(sfd==-1)
		{
			perror("socket PF_NETLINK NETLINK_ROUTE");
			wl_exit();
			return -1;
		}

		memset(&a,0,sizeof(a));
		a.nl_family=AF_NETLINK;
		a.nl_groups=RTMGRP_LINK;

		if(bind(sfd,(struct sockaddr*)&a,sizeof(struct sockaddr_nl)))
		{
			perror("bind PF_NETLINK");
			wl_exit();
			return -1;
		}
	}

	if(checkvers())
	{
		wl_exit();
		return -1;
	}

	return 0;
}

int wl_start(void)
{
	if(pthread_create(&wh,NULL,worker,NULL))
	{
		perror("pthread_create");
		return -1;
	}

	return 0;
}

void wl_stop(void)
{
	uint64_t dummy=1;

	write(efd,&dummy,sizeof(dummy));
	pthread_join(wh,NULL);
}
